# -*- coding: utf-8 -*-
USER_AGENT = (
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36'
)


def make_constant_retry_policy(interval: float):
    def get_interval(_retry_count: int):
        return interval
    return get_interval


def make_linear_retry_policy(start_interval: float, interval_step: float, max_interval: float):
    def get_interval(retry_count: int):
        return min(
            start_interval + (retry_count - 1) * interval_step,
            max_interval
        )
    return get_interval



# -*- coding: utf-8 -*-
import asyncio
import http.cookies
import random
from typing import Optional

import aiohttp

import blivedm
import blivedm.models.web as web_models

# 直播间ID的取值看直播间URL
TEST_ROOM_IDS = [
    12235923,
    14327465,
    21396545,
    21449083,
    23105590,
]

# 这里填一个已登录账号的cookie。不填cookie也可以连接，但是收到弹幕的用户名会打码，UID会变成0
SESSDATA = ''

session: Optional[aiohttp.ClientSession] = None


async def main():
    init_session()
    try:
        await run_single_client()
        await run_multi_clients()
    finally:
        if session is not None:
            await session.close()


def init_session():
    cookies = http.cookies.SimpleCookie()
    cookies['SESSDATA'] = SESSDATA
    cookies['SESSDATA']['domain'] = 'bilibili.com'

    global session
    session = aiohttp.ClientSession()
    session.cookie_jar.update_cookies(cookies)


async def run_single_client():
    """
    演示监听一个直播间
    """
    room_id = random.choice(TEST_ROOM_IDS)
    client = blivedm.BLiveClient(room_id, session=session)
    handler = MyHandler()
    client.set_handler(handler)

    client.start()
    try:
        # 演示5秒后停止
        await asyncio.sleep(5)
        client.stop()

        await client.join()
    finally:
        await client.stop_and_close()


async def run_multi_clients():
    """
    演示同时监听多个直播间
    """
    clients = [blivedm.BLiveClient(room_id, session=session) for room_id in TEST_ROOM_IDS]
    handler = MyHandler()
    for client in clients:
        client.set_handler(handler)
        client.start()

    try:
        await asyncio.gather(*(
            client.join() for client in clients
        ))
    finally:
        await asyncio.gather(*(
            client.stop_and_close() for client in clients
        ))


class MyHandler(blivedm.BaseHandler):
    # # 演示如何添加自定义回调
    # _CMD_CALLBACK_DICT = blivedm.BaseHandler._CMD_CALLBACK_DICT.copy()
    #
    # # 入场消息回调
    # def __interact_word_callback(self, client: blivedm.BLiveClient, command: dict):
    #     print(f"[{client.room_id}] INTERACT_WORD: self_type={type(self).__name__}, room_id={client.room_id},"
    #           f" uname={command['data']['uname']}")
    # _CMD_CALLBACK_DICT['INTERACT_WORD'] = __interact_word_callback  # noqa

    def _on_heartbeat(self, client: blivedm.BLiveClient, message: web_models.HeartbeatMessage):
        print(f'[{client.room_id}] 心跳')

    def _on_danmaku(self, client: blivedm.BLiveClient, message: web_models.DanmakuMessage):
        print(f'[{client.room_id}] {message.uname}：{message.msg}')

    def _on_gift(self, client: blivedm.BLiveClient, message: web_models.GiftMessage):
        print(f'[{client.room_id}] {message.uname} 赠送{message.gift_name}x{message.num}'
              f' （{message.coin_type}瓜子x{message.total_coin}）')

    def _on_buy_guard(self, client: blivedm.BLiveClient, message: web_models.GuardBuyMessage):
        print(f'[{client.room_id}] {message.username} 购买{message.gift_name}')

    def _on_super_chat(self, client: blivedm.BLiveClient, message: web_models.SuperChatMessage):
        print(f'[{client.room_id}] 醒目留言 ¥{message.price} {message.uname}：{message.message}')


if __name__ == '__main__':
    asyncio.run(main())



# -*- coding: utf-8 -*-
import asyncio
import logging
from typing import *

import aiohttp
import yarl

from . import ws_base
from .. import utils

__all__ = (
    'BLiveClient',
)

logger = logging.getLogger('blivedm')

UID_INIT_URL = 'https://api.bilibili.com/x/web-interface/nav'
BUVID_INIT_URL = 'https://data.bilibili.com/v/'
ROOM_INIT_URL = 'https://api.live.bilibili.com/xlive/web-room/v1/index/getInfoByRoom'
DANMAKU_SERVER_CONF_URL = 'https://api.live.bilibili.com/xlive/web-room/v1/index/getDanmuInfo'
DEFAULT_DANMAKU_SERVER_LIST = [
    {'host': 'broadcastlv.chat.bilibili.com', 'port': 2243, 'wss_port': 443, 'ws_port': 2244}
]


class BLiveClient(ws_base.WebSocketClientBase):
    """
    web端客户端

    :param room_id: URL中的房间ID，可以用短ID
    :param uid: B站用户ID，0表示未登录，None表示自动获取
    :param session: cookie、连接池
    :param heartbeat_interval: 发送心跳包的间隔时间（秒）
    """

    def __init__(
        self,
        room_id: int,
        *,
        uid: Optional[int] = None,
        session: Optional[aiohttp.ClientSession] = None,
        heartbeat_interval=30,
    ):
        super().__init__(session, heartbeat_interval)

        self._tmp_room_id = room_id
        """用来init_room的临时房间ID，可以用短ID"""
        self._uid = uid

        # 在调用init_room后初始化的字段
        self._room_owner_uid: Optional[int] = None
        """主播用户ID"""
        self._host_server_list: Optional[List[dict]] = None
        """
        弹幕服务器列表

        `[{host: "tx-bj4-live-comet-04.chat.bilibili.com", port: 2243, wss_port: 443, ws_port: 2244}, ...]`
        """
        self._host_server_token: Optional[str] = None
        """连接弹幕服务器用的token"""

    @property
    def tmp_room_id(self) -> int:
        """
        构造时传进来的room_id参数
        """
        return self._tmp_room_id

    @property
    def room_owner_uid(self) -> Optional[int]:
        """
        主播用户ID，调用init_room后初始化
        """
        return self._room_owner_uid

    @property
    def uid(self) -> Optional[int]:
        """
        当前登录的用户ID，未登录则为0，调用init_room后初始化
        """
        return self._uid

    async def init_room(self):
        """
        初始化连接房间需要的字段

        :return: True代表没有降级，如果需要降级后还可用，重载这个函数返回True
        """
        if self._uid is None:
            if not await self._init_uid():
                logger.warning('room=%d _init_uid() failed', self._tmp_room_id)
                self._uid = 0

        if self._get_buvid() == '':
            if not await self._init_buvid():
                logger.warning('room=%d _init_buvid() failed', self._tmp_room_id)

        res = True
        if not await self._init_room_id_and_owner():
            res = False
            # 失败了则降级
            self._room_id = self._tmp_room_id
            self._room_owner_uid = 0

        if not await self._init_host_server():
            res = False
            # 失败了则降级
            self._host_server_list = DEFAULT_DANMAKU_SERVER_LIST
            self._host_server_token = None
        return res

    async def _init_uid(self):
        cookies = self._session.cookie_jar.filter_cookies(yarl.URL(UID_INIT_URL))
        sessdata_cookie = cookies.get('SESSDATA', None)
        if sessdata_cookie is None or sessdata_cookie.value == '':
            # cookie都没有，不用请求了
            self._uid = 0
            return True

        try:
            async with self._session.get(
                UID_INIT_URL,
                headers={'User-Agent': utils.USER_AGENT},
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _init_uid() failed, status=%d, reason=%s', self._tmp_room_id,
                                   res.status, res.reason)
                    return False
                data = await res.json()
                if data['code'] != 0:
                    if data['code'] == -101:
                        # 未登录
                        self._uid = 0
                        return True
                    logger.warning('room=%d _init_uid() failed, message=%s', self._tmp_room_id,
                                   data['message'])
                    return False

                data = data['data']
                if not data['isLogin']:
                    # 未登录
                    self._uid = 0
                else:
                    self._uid = data['mid']
                return True
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _init_uid() failed:', self._tmp_room_id)
            return False

    def _get_buvid(self):
        cookies = self._session.cookie_jar.filter_cookies(yarl.URL(BUVID_INIT_URL))
        buvid_cookie = cookies.get('buvid3', None)
        if buvid_cookie is None:
            return ''
        return buvid_cookie.value

    async def _init_buvid(self):
        try:
            async with self._session.get(
                BUVID_INIT_URL,
                headers={'User-Agent': utils.USER_AGENT},
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _init_buvid() status error, status=%d, reason=%s',
                                   self._tmp_room_id, res.status, res.reason)
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _init_buvid() exception:', self._tmp_room_id)
        return self._get_buvid() != ''

    async def _init_room_id_and_owner(self):
        try:
            async with self._session.get(
                ROOM_INIT_URL,
                headers={'User-Agent': utils.USER_AGENT},
                params={
                    'room_id': self._tmp_room_id
                },
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _init_room_id_and_owner() failed, status=%d, reason=%s', self._tmp_room_id,
                                   res.status, res.reason)
                    return False
                data = await res.json()
                if data['code'] != 0:
                    logger.warning('room=%d _init_room_id_and_owner() failed, message=%s', self._tmp_room_id,
                                   data['message'])
                    return False
                if not self._parse_room_init(data['data']):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _init_room_id_and_owner() failed:', self._tmp_room_id)
            return False
        return True

    def _parse_room_init(self, data):
        room_info = data['room_info']
        self._room_id = room_info['room_id']
        self._room_owner_uid = room_info['uid']
        return True

    async def _init_host_server(self):
        try:
            async with self._session.get(
                DANMAKU_SERVER_CONF_URL,
                headers={'User-Agent': utils.USER_AGENT},
                params={
                    'id': self._room_id,
                    'type': 0
                },
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _init_host_server() failed, status=%d, reason=%s', self._room_id,
                                   res.status, res.reason)
                    return False
                data = await res.json()
                if data['code'] != 0:
                    logger.warning('room=%d _init_host_server() failed, message=%s', self._room_id, data['message'])
                    return False
                if not self._parse_danmaku_server_conf(data['data']):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _init_host_server() failed:', self._room_id)
            return False
        return True

    def _parse_danmaku_server_conf(self, data):
        self._host_server_list = data['host_list']
        self._host_server_token = data['token']
        if not self._host_server_list:
            logger.warning('room=%d _parse_danmaku_server_conf() failed: host_server_list is empty', self._room_id)
            return False
        return True

    async def _on_before_ws_connect(self, retry_count):
        """
        在每次建立连接之前调用，可以用来初始化房间
        """
        # 重连次数太多则重新init_room，保险
        reinit_period = max(3, len(self._host_server_list or ()))
        if retry_count > 0 and retry_count % reinit_period == 0:
            self._need_init_room = True
        await super()._on_before_ws_connect(retry_count)

    def _get_ws_url(self, retry_count) -> str:
        """
        返回WebSocket连接的URL，可以在这里做故障转移和负载均衡
        """
        host_server = self._host_server_list[retry_count % len(self._host_server_list)]
        return f"wss://{host_server['host']}:{host_server['wss_port']}/sub"

    async def _send_auth(self):
        """
        发送认证包
        """
        auth_params = {
            'uid': self._uid,
            'roomid': self._room_id,
            'protover': 3,
            'platform': 'web',
            'type': 2,
            'buvid': self._get_buvid(),
        }
        if self._host_server_token is not None:
            auth_params['key'] = self._host_server_token
        await self._websocket.send_bytes(self._make_packet(auth_params, ws_base.Operation.AUTH))



# -*- coding: utf-8 -*-
import dataclasses
from typing import *

__all__ = (
    'DanmakuMessage',
    'GiftMessage',
    'GuardBuyMessage',
    'SuperChatMessage',
    'SuperChatDeleteMessage',
    'LikeMessage',
)

# 注释都是复制自官方文档的，看不懂的话问B站
# https://open-live.bilibili.com/document/f9ce25be-312e-1f4a-85fd-fef21f1637f8


@dataclasses.dataclass
class DanmakuMessage:
    """
    弹幕消息
    """

    uname: str = ''
    """用户昵称"""
    uid: int = 0
    """用户UID"""
    uface: str = ''
    """用户头像"""
    timestamp: int = 0
    """弹幕发送时间秒级时间戳"""
    room_id: int = 0
    """弹幕接收的直播间"""
    msg: str = ''
    """弹幕内容"""
    msg_id: str = ''
    """消息唯一id"""
    guard_level: int = 0
    """对应房间大航海等级"""
    fans_medal_wearing_status: bool = False
    """该房间粉丝勋章佩戴情况"""
    fans_medal_name: str = ''
    """粉丝勋章名"""
    fans_medal_level: int = 0
    """对应房间勋章信息"""
    emoji_img_url: str = ''
    """表情包图片地址"""
    dm_type: int = 0
    """弹幕类型 0：普通弹幕 1：表情包弹幕"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uname=data['uname'],
            uid=data['uid'],
            uface=data['uface'],
            timestamp=data['timestamp'],
            room_id=data['room_id'],
            msg=data['msg'],
            msg_id=data['msg_id'],
            guard_level=data['guard_level'],
            fans_medal_wearing_status=data['fans_medal_wearing_status'],
            fans_medal_name=data['fans_medal_name'],
            fans_medal_level=data['fans_medal_level'],
            emoji_img_url=data['emoji_img_url'],
            dm_type=data['dm_type'],
        )


@dataclasses.dataclass
class AnchorInfo:
    """
    主播信息
    """

    uid: int = 0
    """收礼主播uid"""
    uname: str = ''
    """收礼主播昵称"""
    uface: str = ''
    """收礼主播头像"""

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            uid=data['uid'],
            uname=data['uname'],
            uface=data['uface'],
        )


@dataclasses.dataclass
class ComboInfo:
    """
    连击信息
    """

    combo_base_num: int = 0
    """每次连击赠送的道具数量"""
    combo_count: int = 0
    """连击次数"""
    combo_id: str = ''
    """连击id"""
    combo_timeout: int = 0
    """连击有效期秒"""

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            combo_base_num=data['combo_base_num'],
            combo_count=data['combo_count'],
            combo_id=data['combo_id'],
            combo_timeout=data['combo_timeout'],
        )


@dataclasses.dataclass
class GiftMessage:
    """
    礼物消息
    """

    room_id: int = 0
    """房间号"""
    uid: int = 0
    """送礼用户UID"""
    uname: str = ''
    """送礼用户昵称"""
    uface: str = ''
    """送礼用户头像"""
    gift_id: int = 0
    """道具id(盲盒:爆出道具id)"""
    gift_name: str = ''
    """道具名(盲盒:爆出道具名)"""
    gift_num: int = 0
    """赠送道具数量"""
    price: int = 0
    """（礼物单价）支付金额(1000 = 1元 = 10电池),盲盒:爆出道具的价值"""  # 这个B  站文档又不写清楚是单价还是总价
    paid: bool = False
    """是否是付费道具"""
    fans_medal_level: int = 0
    """实际送礼人的勋章信息"""
    fans_medal_name: str = ''
    """粉丝勋章名"""
    fans_medal_wearing_status: bool = False
    """该房间粉丝勋章佩戴情况"""
    guard_level: int = 0
    """大航海等级"""
    timestamp: int = 0
    """收礼时间秒级时间戳"""
    anchor_info: AnchorInfo = dataclasses.field(default_factory=AnchorInfo)
    """主播信息"""
    msg_id: str = ''
    """消息唯一id"""
    gift_icon: str = ''
    """道具icon"""
    combo_gift: bool = False
    """是否是combo道具"""
    combo_info: ComboInfo = dataclasses.field(default_factory=ComboInfo)
    """连击信息"""

    @classmethod
    def from_command(cls, data: dict):
        combo_info = data.get('combo_info', None)
        if combo_info is None:
            combo_info = ComboInfo()
        else:
            combo_info = ComboInfo.from_dict(combo_info)

        return cls(
            room_id=data['room_id'],
            uid=data['uid'],
            uname=data['uname'],
            uface=data['uface'],
            gift_id=data['gift_id'],
            gift_name=data['gift_name'],
            gift_num=data['gift_num'],
            price=data['price'],
            paid=data['paid'],
            fans_medal_level=data['fans_medal_level'],
            fans_medal_name=data['fans_medal_name'],
            fans_medal_wearing_status=data['fans_medal_wearing_status'],
            guard_level=data['guard_level'],
            timestamp=data['timestamp'],
            anchor_info=AnchorInfo.from_dict(data['anchor_info']),
            msg_id=data['msg_id'],
            gift_icon=data['gift_icon'],
            combo_gift=data.get('combo_gift', False),  # 官方的调试工具没发这个字段
            combo_info=combo_info,  # 官方的调试工具没发这个字段
        )


@dataclasses.dataclass
class UserInfo:
    """
    用户信息
    """

    uid: int = 0
    """用户uid"""
    uname: str = ''
    """用户昵称"""
    uface: str = ''
    """用户头像"""

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            uid=data['uid'],
            uname=data['uname'],
            uface=data['uface'],
        )


@dataclasses.dataclass
class GuardBuyMessage:
    """
    上舰消息
    """

    user_info: UserInfo = dataclasses.field(default_factory=UserInfo)
    """用户信息"""
    guard_level: int = 0
    """大航海等级"""
    guard_num: int = 0
    """大航海数量"""
    guard_unit: str = ''
    """大航海单位"""
    fans_medal_level: int = 0
    """粉丝勋章等级"""
    fans_medal_name: str = ''
    """粉丝勋章名"""
    fans_medal_wearing_status: bool = False
    """该房间粉丝勋章佩戴情况"""
    room_id: int = 0
    """房间号"""
    msg_id: str = ''
    """消息唯一id"""
    timestamp: int = 0
    """上舰时间秒级时间戳"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            user_info=UserInfo.from_dict(data['user_info']),
            guard_level=data['guard_level'],
            guard_num=data['guard_num'],
            guard_unit=data['guard_unit'],
            fans_medal_level=data['fans_medal_level'],
            fans_medal_name=data['fans_medal_name'],
            fans_medal_wearing_status=data['fans_medal_wearing_status'],
            room_id=data['room_id'],
            msg_id=data['msg_id'],
            timestamp=data['timestamp'],
        )


@dataclasses.dataclass
class SuperChatMessage:
    """
    醒目留言消息
    """

    room_id: int = 0
    """直播间id"""
    uid: int = 0
    """购买用户UID"""
    uname: str = ''
    """购买的用户昵称"""
    uface: str = ''
    """购买用户头像"""
    message_id: int = 0
    """留言id(风控场景下撤回留言需要)"""
    message: str = ''
    """留言内容"""
    rmb: int = 0
    """支付金额(元)"""
    timestamp: int = 0
    """赠送时间秒级"""
    start_time: int = 0
    """生效开始时间"""
    end_time: int = 0
    """生效结束时间"""
    guard_level: int = 0
    """对应房间大航海等级"""
    fans_medal_level: int = 0
    """对应房间勋章信息"""
    fans_medal_name: str = ''
    """对应房间勋章名字"""
    fans_medal_wearing_status: bool = False
    """该房间粉丝勋章佩戴情况"""
    msg_id: str = ''
    """消息唯一id"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            room_id=data['room_id'],
            uid=data['uid'],
            uname=data['uname'],
            uface=data['uface'],
            message_id=data['message_id'],
            message=data['message'],
            rmb=data['rmb'],
            timestamp=data['timestamp'],
            start_time=data['start_time'],
            end_time=data['end_time'],
            guard_level=data['guard_level'],
            fans_medal_level=data['fans_medal_level'],
            fans_medal_name=data['fans_medal_name'],
            fans_medal_wearing_status=data['fans_medal_wearing_status'],
            msg_id=data['msg_id'],
        )


@dataclasses.dataclass
class SuperChatDeleteMessage:
    """
    删除醒目留言消息
    """

    room_id: int = 0
    """直播间id"""
    message_ids: List[int] = dataclasses.field(default_factory=list)
    """留言id"""
    msg_id: str = ''
    """消息唯一id"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            room_id=data['room_id'],
            message_ids=data['message_ids'],
            msg_id=data['msg_id'],
        )


@dataclasses.dataclass
class LikeMessage:
    """
    点赞消息

    请注意：用户端每分钟触发若干次的情况下只会推送一次该消息
    """

    uname: str = ''
    """用户昵称"""
    uid: int = 0
    """用户UID"""
    uface: str = ''
    """用户头像"""
    timestamp: int = 0
    """时间秒级时间戳"""
    room_id: int = 0
    """发生的直播间"""
    like_text: str = ''
    """点赞文案(“xxx点赞了”)"""
    like_count: int = 0  # 官方文档把这个字段名打错了，这个B文档真是一点都靠不住
    """对单个用户最近2秒的点赞次数聚合"""
    fans_medal_wearing_status: bool = False
    """该房间粉丝勋章佩戴情况"""
    fans_medal_name: str = ''
    """粉丝勋章名"""
    fans_medal_level: int = 0
    """对应房间勋章信息"""
    msg_id: str = ''  # 官方文档表格里没列出这个字段，但是参考JSON里面有
    """消息唯一id"""
    # 还有个guard_level，但官方文档没有出现这个字段，就不添加了

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uname=data['uname'],
            uid=data['uid'],
            uface=data['uface'],
            timestamp=data['timestamp'],
            room_id=data['room_id'],
            like_text=data['like_text'],
            like_count=data['like_count'],
            fans_medal_wearing_status=data['fans_medal_wearing_status'],
            fans_medal_name=data['fans_medal_name'],
            fans_medal_level=data['fans_medal_level'],
            msg_id=data.get('msg_id', ''),  # 官方文档表格里没列出这个字段，但是参考JSON里面有
        )



# -*- coding: utf-8 -*-
import asyncio
import datetime
import hashlib
import hmac
import json
import logging
import random
from typing import *

import aiohttp

from . import ws_base

__all__ = (
    'OpenLiveClient',
)

logger = logging.getLogger('blivedm')

START_URL = 'https://live-open.biliapi.com/v2/app/start'
HEARTBEAT_URL = 'https://live-open.biliapi.com/v2/app/heartbeat'
END_URL = 'https://live-open.biliapi.com/v2/app/end'


class OpenLiveClient(ws_base.WebSocketClientBase):
    """
    开放平台客户端

    文档参考：https://open-live.bilibili.com/document/

    :param access_key_id: 在开放平台申请的access_key_id
    :param access_key_secret: 在开放平台申请的access_key_secret
    :param app_id: 在开放平台创建的项目ID
    :param room_owner_auth_code: 主播身份码
    :param session: cookie、连接池
    :param heartbeat_interval: 发送连接心跳包的间隔时间（秒）
    :param game_heartbeat_interval: 发送项目心跳包的间隔时间（秒）
    """

    def __init__(
        self,
        access_key_id: str,
        access_key_secret: str,
        app_id: int,
        room_owner_auth_code: str,
        *,
        session: Optional[aiohttp.ClientSession] = None,
        heartbeat_interval=30,
        game_heartbeat_interval=20,
    ):
        super().__init__(session, heartbeat_interval)

        self._access_key_id = access_key_id
        self._access_key_secret = access_key_secret
        self._app_id = app_id
        self._room_owner_auth_code = room_owner_auth_code
        self._game_heartbeat_interval = game_heartbeat_interval

        # 在调用init_room后初始化的字段
        self._room_owner_uid: Optional[int] = None
        """主播用户ID"""
        self._host_server_url_list: Optional[List[str]] = []
        """弹幕服务器URL列表"""
        self._auth_body: Optional[str] = None
        """连接弹幕服务器用的认证包内容"""
        self._game_id: Optional[str] = None
        """项目场次ID"""

        # 在运行时初始化的字段
        self._game_heartbeat_timer_handle: Optional[asyncio.TimerHandle] = None
        """发项目心跳包定时器的handle"""

    @property
    def room_owner_uid(self) -> Optional[int]:
        """
        主播用户ID，调用init_room后初始化
        """
        return self._room_owner_uid

    @property
    def room_owner_auth_code(self):
        """
        主播身份码
        """
        return self._room_owner_auth_code

    @property
    def app_id(self):
        """
        在开放平台创建的项目ID
        """
        return self._app_id

    @property
    def game_id(self) -> Optional[str]:
        """
        项目场次ID，调用init_room后初始化
        """
        return self._game_id

    async def close(self):
        """
        释放本客户端的资源，调用后本客户端将不可用
        """
        if self.is_running:
            logger.warning('room=%s is calling close(), but client is running', self.room_id)

        if self._game_heartbeat_timer_handle is not None:
            self._game_heartbeat_timer_handle.cancel()
            self._game_heartbeat_timer_handle = None
        await self._end_game()

        await super().close()

    def _request_open_live(self, url, body: dict):
        body_bytes = json.dumps(body).encode('utf-8')
        headers = {
            'x-bili-accesskeyid': self._access_key_id,
            'x-bili-content-md5': hashlib.md5(body_bytes).hexdigest(),
            'x-bili-signature-method': 'HMAC-SHA256',
            'x-bili-signature-nonce': str(random.randint(0, 999999999)),
            'x-bili-signature-version': '1.0',
            'x-bili-timestamp': str(int(datetime.datetime.now().timestamp())),
        }

        str_to_sign = '\n'.join(
            f'{key}:{value}'
            for key, value in headers.items()
        )
        signature = hmac.new(
            self._access_key_secret.encode('utf-8'), str_to_sign.encode('utf-8'), hashlib.sha256
        ).hexdigest()
        headers['Authorization'] = signature

        headers['Content-Type'] = 'application/json'
        headers['Accept'] = 'application/json'
        return self._session.post(url, headers=headers, data=body_bytes)

    async def init_room(self):
        """
        开启项目，并初始化连接房间需要的字段

        :return: 是否成功
        """
        if not await self._start_game():
            return False

        if self._game_id != '' and self._game_heartbeat_timer_handle is None:
            self._game_heartbeat_timer_handle = asyncio.get_running_loop().call_later(
                self._game_heartbeat_interval, self._on_send_game_heartbeat
            )
        return True

    async def _start_game(self):
        try:
            async with self._request_open_live(
                START_URL,
                {'code': self._room_owner_auth_code, 'app_id': self._app_id}
            ) as res:
                if res.status != 200:
                    logger.warning('_start_game() failed, status=%d, reason=%s', res.status, res.reason)
                    return False
                data = await res.json()
                if data['code'] != 0:
                    logger.warning('_start_game() failed, code=%d, message=%s, request_id=%s',
                                   data['code'], data['message'], data['request_id'])
                    return False
                if not self._parse_start_game(data['data']):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('_start_game() failed:')
            return False
        return True

    def _parse_start_game(self, data):
        self._game_id = data['game_info']['game_id']
        websocket_info = data['websocket_info']
        self._auth_body = websocket_info['auth_body']
        self._host_server_url_list = websocket_info['wss_link']
        anchor_info = data['anchor_info']
        self._room_id = anchor_info['room_id']
        self._room_owner_uid = anchor_info['uid']
        return True

    async def _end_game(self):
        """
        关闭项目。建议关闭客户端时保证调用到这个函数（close会调用），否则可能短时间内无法重复连接同一个房间
        """
        if self._game_id in (None, ''):
            return True

        try:
            async with self._request_open_live(
                END_URL,
                {'app_id': self._app_id, 'game_id': self._game_id}
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _end_game() failed, status=%d, reason=%s',
                                   self._room_id, res.status, res.reason)
                    return False
                data = await res.json()
                code = data['code']
                if code != 0:
                    if code in (7000, 7003):
                        # 项目已经关闭了也算成功
                        return True

                    logger.warning('room=%d _end_game() failed, code=%d, message=%s, request_id=%s',
                                   self._room_id, code, data['message'], data['request_id'])
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _end_game() failed:', self._room_id)
            return False
        return True

    def _on_send_game_heartbeat(self):
        """
        定时发送项目心跳包的回调
        """
        self._game_heartbeat_timer_handle = asyncio.get_running_loop().call_later(
            self._game_heartbeat_interval, self._on_send_game_heartbeat
        )
        asyncio.create_task(self._send_game_heartbeat())

    async def _send_game_heartbeat(self):
        """
        发送项目心跳包
        """
        if self._game_id in (None, ''):
            logger.warning('game=%d _send_game_heartbeat() failed, game_id not found', self._game_id)
            return False

        try:
            # 保存一下，防止await之后game_id改变
            game_id = self._game_id
            async with self._request_open_live(
                HEARTBEAT_URL,
                {'game_id': game_id}
            ) as res:
                if res.status != 200:
                    logger.warning('room=%d _send_game_heartbeat() failed, status=%d, reason=%s',
                                   self._room_id, res.status, res.reason)
                    return False
                data = await res.json()
                code = data['code']
                if code != 0:
                    logger.warning('room=%d _send_game_heartbeat() failed, code=%d, message=%s, request_id=%s',
                                   self._room_id, code, data['message'], data['request_id'])

                    if code == 7003 and self._game_id == game_id:
                        # 项目异常关闭，可能是心跳超时，需要重新开启项目
                        self._need_init_room = True
                        if self._websocket is not None and not self._websocket.closed:
                            await self._websocket.close()

                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception('room=%d _send_game_heartbeat() failed:', self._room_id)
            return False
        return True

    async def _on_before_ws_connect(self, retry_count):
        """
        在每次建立连接之前调用，可以用来初始化房间
        """
        # 重连次数太多则重新init_room，保险
        reinit_period = max(3, len(self._host_server_url_list or ()))
        if retry_count > 0 and retry_count % reinit_period == 0:
            self._need_init_room = True
        await super()._on_before_ws_connect(retry_count)

    def _get_ws_url(self, retry_count) -> str:
        """
        返回WebSocket连接的URL，可以在这里做故障转移和负载均衡
        """
        return self._host_server_url_list[retry_count % len(self._host_server_url_list)]

    async def _send_auth(self):
        """
        发送认证包
        """
        await self._websocket.send_bytes(self._make_packet(self._auth_body, ws_base.Operation.AUTH))



# -*- coding: utf-8 -*-
import logging
from typing import *

from .clients import ws_base
from .models import web as web_models, open_live as open_models

__all__ = (
    'HandlerInterface',
    'BaseHandler',
)

logger = logging.getLogger('blivedm')

logged_unknown_cmds = {
    'COMBO_SEND',
    'ENTRY_EFFECT',
    'HOT_RANK_CHANGED',
    'HOT_RANK_CHANGED_V2',
    # 'INTERACT_WORD', # 进入直播间
    # 'LIVE',
    'LIVE_INTERACTIVE_GAME',
    'NOTICE_MSG',
    'ONLINE_RANK_COUNT',
    'ONLINE_RANK_TOP3',
    'ONLINE_RANK_V2',
    'PK_BATTLE_END',
    'PK_BATTLE_ENTRANCE',
    'PK_BATTLE_FINAL_PROCESS',
    'PK_BATTLE_MATCH_TIMEOUT',
    'PK_BATTLE_PRE',
    'PK_BATTLE_PRE_NEW',
    'PK_BATTLE_PROCESS',
    'PK_BATTLE_PROCESS_NEW',
    'PK_BATTLE_SETTLE',
    'PK_BATTLE_SETTLE_NEW',
    'PK_BATTLE_SETTLE_USER',
    'PK_BATTLE_SETTLE_V2',
    'PK_BATTLE_START',
    'PK_BATTLE_START_NEW',
    'PK_BATTLE_VIDEO_PUNISH_BEGIN',
    'PK_BATTLE_VIDEO_PUNISH_END',
    # 'PREPARING',
    'ROOM_REAL_TIME_MESSAGE_UPDATE',
    'STOP_LIVE_ROOM_LIST',
    'SUPER_CHAT_MESSAGE_JPN',
    'WIDGET_BANNER',
    'WATCHED_CHANGE',
    'LIVE_MULTI_VIEW_CHANGE',
    'COMMON_NOTICE_DANMAKU', # BLS秋季限时任务：任务即将结束，抓紧完成获取10元红包奖励吧！未完成任务进度将无法获得奖励 https://st.openbayes.net/organizations/openbayes/issues/732/events/1fcc24e45ef34148b3193fb676dffc5a/?project=13&query=is%3Aunresolved
    'DANMU_AGGREGATION', # 老板大气！点点红包抽礼物, https://st.openbayes.net/organizations/openbayes/issues/732/events/aafcb5e6c02b45af8de99a26175e25fe/?project=13&query=is%3Aunresolved
    'HOT_RANK_SETTLEMENT', # 恭喜主播 <% 阿梓从小就很可爱 %> 荣登限时热门榜虚拟主播榜top10! 即将获得热门流量推荐哦！
    'HOT_RANK_SETTLEMENT_V2', # 恭喜主播 <% 老实憨厚的笑笑 %> 荣登限时热门榜户外榜榜首! 即将获得热门流量推荐哦！
}
"""已打日志的未知cmd"""


class HandlerInterface:
    """
    直播消息处理器接口
    """

    def handle(self, client: ws_base.WebSocketClientBase, command: dict):
        raise NotImplementedError

    def on_client_stopped(self, client: ws_base.WebSocketClientBase, exception: Optional[Exception]):
        """
        当客户端停止时调用。可以在这里close或者重新start
        """


def _make_msg_callback(method_name, message_cls):
    def callback(self: 'BaseHandler', client: ws_base.WebSocketClientBase, command: dict):
        method = getattr(self, method_name)
        return method(client, message_cls.from_command(command['data']))
    return callback

def _make_raw_msg_callback(method_name, message_cls, raw = False):
    def callback(self: 'BaseHandler', client: ws_base.WebSocketClientBase, command: dict):
        method = getattr(self, method_name)
        return method(client, message_cls.from_command(
            command if raw else command['data']
        ))
    return callback


class BaseHandler(HandlerInterface):
    """
    一个简单的消息处理器实现，带消息分发和消息类型转换。继承并重写_on_xxx方法即可实现自己的处理器
    """

    def __danmu_msg_callback(self, client: ws_base.WebSocketClientBase, command: dict):
        return self._on_danmaku(
            client, web_models.DanmakuMessage.from_command(command['info'], command.get('dm_v2', ''))
        )

    _CMD_CALLBACK_DICT: Dict[
        str,
        Optional[Callable[
            ['BaseHandler', ws_base.WebSocketClientBase, dict],
            Any
        ]]
    ] = {
        # 收到心跳包，这是blivedm自造的消息，原本的心跳包格式不一样
        '_HEARTBEAT': _make_msg_callback('_on_heartbeat', web_models.HeartbeatMessage),
        # 收到弹幕
        # go-common\app\service\live\live-dm\service\v1\send.go
        'DANMU_MSG': __danmu_msg_callback,
        # 有人送礼
        'SEND_GIFT': _make_msg_callback('_on_gift', web_models.GiftMessage),
        # 有人上舰
        'GUARD_BUY': _make_msg_callback('_on_buy_guard', web_models.GuardBuyMessage),
        # 醒目留言
        'SUPER_CHAT_MESSAGE': _make_msg_callback('_on_super_chat', web_models.SuperChatMessage),
        # 删除醒目留言
        'SUPER_CHAT_MESSAGE_DELETE': _make_msg_callback('_on_super_chat_delete', web_models.SuperChatDeleteMessage),

        # eop-blive 自定义事件
        'INTERACT_WORD': _make_raw_msg_callback('_on_interact_word', web_models.InteractWordMessage),
        'USER_TOAST_MSG': _make_raw_msg_callback('_on_user_toast', web_models.UserToastMessage),
        'USER_VIRTUAL_MVP': _make_raw_msg_callback('_on_virtual_mvp', web_models.UserVirtualMvp),
        'WARNING': _make_raw_msg_callback('_on_warning', web_models.WarningMessage, True),
        'CUT_OFF': _make_raw_msg_callback('_on_cut_off', web_models.CutOffMessage, True),
        'ROOM_SILENT_ON': _make_raw_msg_callback('_on_room_silent_on', web_models.RoomSilentOnMessage),
        'ROOM_SILENT_OFF': _make_raw_msg_callback('_on_room_silent_off', web_models.RoomSilentOffMessage),
        'ROOM_ADMINS': _make_raw_msg_callback('_on_room_admins', web_models.RoomAdminsMessage, True),
        # yes this is the lowercased event from ws 😅
        'room_admin_entrance': _make_raw_msg_callback('_on_room_admin_entrance', web_models.RoomAdminEntranceMessage, True),
        'ROOM_ADMIN_REVOKE': _make_raw_msg_callback('_on_room_admin_revoke', web_models.RoomAdminRevokeMessage, True),
        'ROOM_BLOCK_MSG': _make_raw_msg_callback('_on_room_block', web_models.RoomBlockMessage),

        #
        # 开放平台消息
        #

        # 收到弹幕
        'LIVE_OPEN_PLATFORM_DM': _make_msg_callback('_on_open_live_danmaku', open_models.DanmakuMessage),
        # 有人送礼
        'LIVE_OPEN_PLATFORM_SEND_GIFT': _make_msg_callback('_on_open_live_gift', open_models.GiftMessage),
        # 有人上舰
        'LIVE_OPEN_PLATFORM_GUARD': _make_msg_callback('_on_open_live_buy_guard', open_models.GuardBuyMessage),
        # 醒目留言
        'LIVE_OPEN_PLATFORM_SUPER_CHAT': _make_msg_callback('_on_open_live_super_chat', open_models.SuperChatMessage),
        # 删除醒目留言
        'LIVE_OPEN_PLATFORM_SUPER_CHAT_DEL': _make_msg_callback(
            '_on_open_live_super_chat_delete', open_models.SuperChatDeleteMessage
        ),
        # 点赞
        'LIVE_OPEN_PLATFORM_LIKE': _make_msg_callback('_on_open_live_like', open_models.LikeMessage),
    }
    """cmd -> 处理回调"""

    def handle(self, client: ws_base.WebSocketClientBase, command: dict):
        cmd = command.get('cmd', '')
        pos = cmd.find(':')  # 2019-5-29 B站弹幕升级新增了参数
        if pos != -1:
            cmd = cmd[:pos]

        if cmd not in self._CMD_CALLBACK_DICT:
            # 只有第一次遇到未知cmd时打日志
            if cmd not in logged_unknown_cmds:
                logger.warning('room=%d unknown cmd=%s, command=%s', client.room_id, cmd, command)
                logged_unknown_cmds.add(cmd)
            return

        callback = self._CMD_CALLBACK_DICT[cmd]
        if callback is not None:
            callback(self, client, command)

    def _on_heartbeat(self, client: ws_base.WebSocketClientBase, message: web_models.HeartbeatMessage):
        """
        收到心跳包
        """

    def _on_danmaku(self, client: ws_base.WebSocketClientBase, message: web_models.DanmakuMessage):
        """
        收到弹幕
        """

    def _on_gift(self, client: ws_base.WebSocketClientBase, message: web_models.GiftMessage):
        """
        收到礼物
        """

    def _on_buy_guard(self, client: ws_base.WebSocketClientBase, message: web_models.GuardBuyMessage):
        """
        有人上舰
        """

    def _on_super_chat(self, client: ws_base.WebSocketClientBase, message: web_models.SuperChatMessage):
        """
        醒目留言
        """

    def _on_super_chat_delete(
        self, client: ws_base.WebSocketClientBase, message: web_models.SuperChatDeleteMessage
    ):
        """
        删除醒目留言
        """

    # eop-blive 自定义事件
    def _on_interact_word(
        self, client: ws_base.WebSocketClientBase, message: web_models.InteractWordMessage
    ):
        """
        _on_interact_word
        """
    def _on_user_toast(
        self, client: ws_base.WebSocketClientBase, message: web_models.UserToastMessage
    ):
        """
        _on_user_toast
        """
    def _on_virtual_mvp(
        self, client: ws_base.WebSocketClientBase, message: web_models.UserVirtualMvp
    ):
        """
        _on_virtual_mvp
        """
    def _on_warning(
        self, client: ws_base.WebSocketClientBase, message: web_models.WarningMessage
    ):
        """
        _on_warning
        """
    def _on_cut_off(
        self, client: ws_base.WebSocketClientBase, message: web_models.CutOffMessage
    ):
        """
        _on_cut_off
        """
    def _on_room_silent_on(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomSilentOnMessage
    ):
        """
        _on_room_silent_on
        """
    def _on_room_silent_off(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomSilentOffMessage
    ):
        """
        _on_room_silent_off
        """
    def _on_room_admins(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomAdminsMessage
    ):
        """
        _on_room_admins
        """
    def _on_room_admin_entrance(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomAdminEntranceMessage
    ):
        """
        _on_room_admin_entrance
        """
    def _on_room_admin_revoke(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomAdminRevokeMessage
    ):
        """
        _on_room_admin_revoke
        """
    def _on_room_block(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomBlockMessage
    ):
        """
        _on_room_block
        """

    #
    # 开放平台消息
    #

    def _on_open_live_danmaku(self, client: ws_base.WebSocketClientBase, message: open_models.DanmakuMessage):
        """
        收到弹幕
        """

    def _on_open_live_gift(self, client: ws_base.WebSocketClientBase, message: open_models.GiftMessage):
        """
        收到礼物
        """

    def _on_open_live_buy_guard(self, client: ws_base.WebSocketClientBase, message: open_models.GuardBuyMessage):
        """
        有人上舰
        """

    def _on_open_live_super_chat(
        self, client: ws_base.WebSocketClientBase, message: open_models.SuperChatMessage
    ):
        """
        醒目留言
        """

    def _on_open_live_super_chat_delete(
        self, client: ws_base.WebSocketClientBase, message: open_models.SuperChatDeleteMessage
    ):
        """
        删除醒目留言
        """

    def _on_open_live_like(self, client: ws_base.WebSocketClientBase, message: open_models.LikeMessage):
        """
        点赞
        """



# -*- coding: utf-8 -*-
import dataclasses
import enum
from typing import *

import pure_protobuf.annotations as pb_anno
import pure_protobuf.message as pb_msg

try:
    Annotated
except NameError:
    from typing_extensions import Annotated  # Python < 3.9


@dataclasses.dataclass
class SimpleUser(pb_msg.BaseMessage):
    face: Annotated[str, pb_anno.Field(4)] = ''


@dataclasses.dataclass
class SimpleDm(pb_msg.BaseMessage):
    user: Annotated[SimpleUser, pb_anno.Field(20)] = dataclasses.field(default_factory=SimpleUser)


#
# 以下代码是预防以后全量使用Protobuf协议
#

class BizScene(enum.IntEnum):
    None_ = 0
    Lottery = 1
    Survive = 2
    VoiceConn = 3
    PlayBack = 4
    Vote = 5


@dataclasses.dataclass
class Bubble(pb_msg.BaseMessage):
    id: Annotated[int, pb_anno.Field(1)] = 0
    color: Annotated[str, pb_anno.Field(2)] = ''
    id_v2: Annotated[int, pb_anno.Field(3)] = 0


class DmType(enum.IntEnum):
    Normal = 0
    Emoticon = 1
    Voice = 2


@dataclasses.dataclass
class Emoticon(pb_msg.BaseMessage):
    unique: Annotated[str, pb_anno.Field(1)] = ''
    url: Annotated[str, pb_anno.Field(2)] = ''
    is_dynamic: Annotated[bool, pb_anno.Field(3)] = False
    in_player_area: Annotated[int, pb_anno.Field(4)] = 0
    bulge_display: Annotated[int, pb_anno.Field(5)] = 0
    height: Annotated[int, pb_anno.Field(6)] = 0
    width: Annotated[int, pb_anno.Field(7)] = 0


# pure_protobuf不支持map的临时解决方案
@dataclasses.dataclass
class EmoticonMapEntry(pb_msg.BaseMessage):
    key: Annotated[str, pb_anno.Field(1)] = ''
    value: Annotated[Emoticon, pb_anno.Field(2)] = dataclasses.field(default_factory=Emoticon)


@dataclasses.dataclass
class Voice(pb_msg.BaseMessage):
    url: Annotated[str, pb_anno.Field(1)] = ''
    file_format: Annotated[str, pb_anno.Field(2)] = ''
    text: Annotated[str, pb_anno.Field(3)] = ''
    file_duration: Annotated[int, pb_anno.Field(4)] = 0
    file_id: Annotated[str, pb_anno.Field(5)] = ''


@dataclasses.dataclass
class Aggregation(pb_msg.BaseMessage):
    is_aggregation: Annotated[bool, pb_anno.Field(1)] = False
    activity_source: Annotated[int, pb_anno.Field(2)] = 0
    activity_identity: Annotated[str, pb_anno.Field(3)] = ''
    not_show: Annotated[int, pb_anno.Field(4)] = 0


@dataclasses.dataclass
class Check(pb_msg.BaseMessage):
    token: Annotated[str, pb_anno.Field(1)] = ''
    ts: Annotated[int, pb_anno.Field(2)] = 0


@dataclasses.dataclass
class Medal(pb_msg.BaseMessage):
    level: Annotated[int, pb_anno.Field(1)] = 0
    name: Annotated[str, pb_anno.Field(2)] = ''
    special: Annotated[str, pb_anno.Field(3)] = ''
    color: Annotated[int, pb_anno.Field(4)] = 0
    icon_id: Annotated[int, pb_anno.Field(5)] = 0
    border_color: Annotated[int, pb_anno.Field(6)] = 0
    gradient_start_color: Annotated[int, pb_anno.Field(7)] = 0
    gradient_end_color: Annotated[int, pb_anno.Field(8)] = 0
    privilege: Annotated[int, pb_anno.Field(9)] = 0
    light: Annotated[int, pb_anno.Field(10)] = 0


@dataclasses.dataclass
class UserLevel(pb_msg.BaseMessage):
    level: Annotated[int, pb_anno.Field(1)] = 0
    color: Annotated[int, pb_anno.Field(2)] = 0
    rank: Annotated[str, pb_anno.Field(3)] = ''
    online_rank: Annotated[int, pb_anno.Field(4)] = 0


@dataclasses.dataclass
class Title(pb_msg.BaseMessage):
    title: Annotated[str, pb_anno.Field(1)] = ''
    old_title: Annotated[str, pb_anno.Field(2)] = ''


@dataclasses.dataclass
class Identify(pb_msg.BaseMessage):
    beginning_url: Annotated[str, pb_anno.Field(1)] = ''
    ending_url: Annotated[str, pb_anno.Field(2)] = ''
    jump_to_url: Annotated[str, pb_anno.Field(3)] = ''


@dataclasses.dataclass
class Wealth(pb_msg.BaseMessage):
    level: Annotated[int, pb_anno.Field(1)] = 0


@dataclasses.dataclass
class User(pb_msg.BaseMessage):
    uid: Annotated[int, pb_anno.Field(1)] = 0
    name: Annotated[str, pb_anno.Field(2)] = ''
    name_color: Annotated[str, pb_anno.Field(3)] = ''
    face: Annotated[str, pb_anno.Field(4)] = ''
    vip: Annotated[int, pb_anno.Field(5)] = 0
    svip: Annotated[int, pb_anno.Field(6)] = 0
    rank: Annotated[int, pb_anno.Field(7)] = 0
    mobile_verify: Annotated[int, pb_anno.Field(8)] = 0
    lpl_status: Annotated[int, pb_anno.Field(9)] = 0
    attr: Annotated[int, pb_anno.Field(10)] = 0
    medal: Annotated[Medal, pb_anno.Field(11)] = dataclasses.field(default_factory=Medal)
    level: Annotated[UserLevel, pb_anno.Field(12)] = dataclasses.field(default_factory=UserLevel)
    title: Annotated[Title, pb_anno.Field(13)] = dataclasses.field(default_factory=Title)
    identify: Annotated[Identify, pb_anno.Field(14)] = dataclasses.field(default_factory=Identify)
    wealth: Annotated[Wealth, pb_anno.Field(15)] = dataclasses.field(default_factory=Wealth)


@dataclasses.dataclass
class Room(pb_msg.BaseMessage):
    uid: Annotated[int, pb_anno.Field(1)] = 0
    name: Annotated[str, pb_anno.Field(2)] = ''


@dataclasses.dataclass
class Prefix(pb_msg.BaseMessage):
    type: Annotated[int, pb_anno.Field(1)] = 0
    resource: Annotated[str, pb_anno.Field(2)] = ''


@dataclasses.dataclass
class Icon(pb_msg.BaseMessage):
    prefix: Annotated[Prefix, pb_anno.Field(1)] = dataclasses.field(default_factory=Prefix)


@dataclasses.dataclass
class Dm(pb_msg.BaseMessage):
    id_str: Annotated[str, pb_anno.Field(1)] = ''
    mode: Annotated[int, pb_anno.Field(2)] = 0
    fontsize: Annotated[int, pb_anno.Field(3)] = 0
    color: Annotated[pb_anno.uint, pb_anno.Field(4)] = 0
    mid_hash: Annotated[str, pb_anno.Field(5)] = ''
    content: Annotated[str, pb_anno.Field(6)] = ''
    ctime: Annotated[int, pb_anno.Field(7)] = 0
    weight: Annotated[int, pb_anno.Field(8)] = 0
    rnd: Annotated[int, pb_anno.Field(9)] = 0
    attr: Annotated[int, pb_anno.Field(10)] = 0
    # 为了防止加新枚举后不兼容，还是用int了
    # biz_scene: Annotated[BizScene, pb_anno.Field(11)] = BizScene.None_
    biz_scene: Annotated[int, pb_anno.Field(11)] = 0
    bubble: Annotated[Bubble, pb_anno.Field(12)] = dataclasses.field(default_factory=Bubble)
    # dm_type: Annotated[DmType, pb_anno.Field(13)] = DmType.Normal
    dm_type: Annotated[int, pb_anno.Field(13)] = 0
    emoticons: Annotated[List[EmoticonMapEntry], pb_anno.Field(14)] = dataclasses.field(default_factory=list)
    voice: Annotated[Voice, pb_anno.Field(15)] = dataclasses.field(default_factory=Voice)
    animation: Annotated[str, pb_anno.Field(16)] = ''
    aggregation: Annotated[Aggregation, pb_anno.Field(17)] = dataclasses.field(default_factory=Aggregation)
    send_from_me: Annotated[bool, pb_anno.Field(18)] = False
    check: Annotated[Check, pb_anno.Field(19)] = dataclasses.field(default_factory=Check)
    user: Annotated[User, pb_anno.Field(20)] = dataclasses.field(default_factory=User)
    room: Annotated[Room, pb_anno.Field(21)] = dataclasses.field(default_factory=Room)
    icon: Annotated[Icon, pb_anno.Field(22)] = dataclasses.field(default_factory=Icon)



# -*- coding: utf-8 -*-
import base64
import binascii
import dataclasses
import json
from typing import *

from . import pb

__all__ = (
    'HeartbeatMessage',
    'DanmakuMessage',
    'GiftMessage',
    'GuardBuyMessage',
    'SuperChatMessage',
    'SuperChatDeleteMessage',

    # custom cmds
    'InteractWordMessage',
    'UserToastMessage',
    'UserVirtualMvp',
    'WarningMessage',
    'CutOffMessage',
    'RoomSilentOnMessage',
    'RoomSilentOffMessage',
    'RoomAdminsMessage',
    'RoomAdminEntranceMessage',
    'RoomAdminRevokeMessage',
    'RoomBlockMessage',
)


@dataclasses.dataclass
class HeartbeatMessage:
    """
    心跳消息
    """

    popularity: int = 0
    """人气值，已废弃"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            popularity=data['popularity'],
        )


@dataclasses.dataclass
class DanmakuMessage:
    """
    弹幕消息
    """

    mode: int = 0
    """弹幕显示模式（滚动、顶部、底部）"""
    font_size: int = 0
    """字体尺寸"""
    color: int = 0
    """颜色"""
    timestamp: int = 0
    """时间戳（毫秒）"""
    rnd: int = 0
    """随机数，前端叫作弹幕ID，可能是去重用的"""
    uid_crc32: str = ''
    """用户ID文本的CRC32"""
    msg_type: int = 0
    """是否礼物弹幕（节奏风暴）"""
    bubble: int = 0
    """右侧评论栏气泡"""
    dm_type: int = 0
    """弹幕类型，0文本，1表情，2语音"""
    emoticon_options: Union[dict, str] = ''
    """表情参数"""
    voice_config: Union[dict, str] = ''
    """语音参数"""
    mode_info: dict = dataclasses.field(default_factory=dict)
    """一些附加参数"""

    msg: str = ''
    """弹幕内容"""

    uid: int = 0
    """用户ID"""
    uname: str = ''
    """用户名"""
    face: str = ''
    """用户头像URL"""
    admin: int = 0
    """是否房管"""
    vip: int = 0
    """是否月费老爷"""
    svip: int = 0
    """是否年费老爷"""
    urank: int = 0
    """用户身份，用来判断是否正式会员，猜测非正式会员为5000，正式会员为10000"""
    mobile_verify: int = 0
    """是否绑定手机"""
    uname_color: str = ''
    """用户名颜色"""

    medal_level: str = ''
    """勋章等级"""
    medal_name: str = ''
    """勋章名"""
    runame: str = ''
    """勋章房间主播名"""
    medal_room_id: int = 0
    """勋章房间ID"""
    mcolor: int = 0
    """勋章颜色"""
    special_medal: str = ''
    """特殊勋章"""

    user_level: int = 0
    """用户等级"""
    ulevel_color: int = 0
    """用户等级颜色"""
    ulevel_rank: str = ''
    """用户等级排名，>50000时为'>50000'"""

    old_title: str = ''
    """旧头衔"""
    title: str = ''
    """头衔"""

    privilege_type: int = 0
    """舰队类型，0非舰队，1总督，2提督，3舰长"""

    @classmethod
    def from_command(cls, info: list, dm_v2=''):
        proto: Optional[pb.SimpleDm] = None
        if dm_v2 != '':
            try:
                proto = pb.SimpleDm.loads(base64.b64decode(dm_v2))
            except (binascii.Error, KeyError, TypeError, ValueError):
                pass
        if proto is not None:
            face = proto.user.face
        else:
            face = ''

        if len(info[3]) != 0:
            medal_level = info[3][0]
            medal_name = info[3][1]
            runame = info[3][2]
            room_id = info[3][3]
            mcolor = info[3][4]
            special_medal = info[3][5]
        else:
            medal_level = 0
            medal_name = ''
            runame = ''
            room_id = 0
            mcolor = 0
            special_medal = 0

        return cls(
            mode=info[0][1],
            font_size=info[0][2],
            color=info[0][3],
            timestamp=info[0][4],
            rnd=info[0][5],
            uid_crc32=info[0][7],
            msg_type=info[0][9],
            bubble=info[0][10],
            dm_type=info[0][12],
            emoticon_options=info[0][13],
            voice_config=info[0][14],
            mode_info=info[0][15],

            msg=info[1],

            uid=info[2][0],
            uname=info[2][1],
            face=face,
            admin=info[2][2],
            vip=info[2][3],
            svip=info[2][4],
            urank=info[2][5],
            mobile_verify=info[2][6],
            uname_color=info[2][7],

            medal_level=medal_level,
            medal_name=medal_name,
            runame=runame,
            medal_room_id=room_id,
            mcolor=mcolor,
            special_medal=special_medal,

            user_level=info[4][0],
            ulevel_color=info[4][2],
            ulevel_rank=info[4][3],

            old_title=info[5][0],
            title=info[5][1],

            privilege_type=info[7],
        )

    @property
    def emoticon_options_dict(self) -> dict:
        """
        示例：
        {'bulge_display': 0, 'emoticon_unique': 'official_13', 'height': 60, 'in_player_area': 1, 'is_dynamic': 1,
         'url': 'https://i0.hdslb.com/bfs/live/a98e35996545509188fe4d24bd1a56518ea5af48.png', 'width': 183}
        """
        if isinstance(self.emoticon_options, dict):
            return self.emoticon_options
        try:
            return json.loads(self.emoticon_options)
        except (json.JSONDecodeError, TypeError):
            return {}

    @property
    def voice_config_dict(self) -> dict:
        """
        示例：
        {'voice_url': 'https%3A%2F%2Fboss.hdslb.com%2Flive-dm-voice%2Fb5b26e48b556915cbf3312a59d3bb2561627725945.wav
         %3FX-Amz-Algorithm%3DAWS4-HMAC-SHA256%26X-Amz-Credential%3D2663ba902868f12f%252F20210731%252Fshjd%252Fs3%25
         2Faws4_request%26X-Amz-Date%3D20210731T100545Z%26X-Amz-Expires%3D600000%26X-Amz-SignedHeaders%3Dhost%26
         X-Amz-Signature%3D114e7cb5ac91c72e231c26d8ca211e53914722f36309b861a6409ffb20f07ab8',
         'file_format': 'wav', 'text': '汤，下午好。', 'file_duration': 1}
        """
        if isinstance(self.voice_config, dict):
            return self.voice_config
        try:
            return json.loads(self.voice_config)
        except (json.JSONDecodeError, TypeError):
            return {}


@dataclasses.dataclass
class GiftMessage:
    """
    礼物消息
    """

    gift_name: str = ''
    """礼物名"""
    num: int = 0
    """数量"""
    uname: str = ''
    """用户名"""
    face: str = ''
    """用户头像URL"""
    guard_level: int = 0
    """舰队等级，0非舰队，1总督，2提督，3舰长"""
    uid: int = 0
    """用户ID"""
    timestamp: int = 0
    """时间戳"""
    gift_id: int = 0
    """礼物ID"""
    gift_type: int = 0
    """礼物类型（未知）"""
    action: str = ''
    """目前遇到的有'喂食'、'赠送'"""
    price: int = 0
    """礼物单价瓜子数"""
    rnd: str = ''
    """随机数，可能是去重用的。有时是时间戳+去重ID，有时是UUID"""
    coin_type: str = ''
    """瓜子类型，'silver'或'gold'，1000金瓜子 = 1元"""
    total_coin: int = 0
    """总瓜子数"""
    tid: str = ''
    """可能是事务ID，有时和rnd相同"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            gift_name=data['giftName'],
            num=data['num'],
            uname=data['uname'],
            face=data['face'],
            guard_level=data['guard_level'],
            uid=data['uid'],
            timestamp=data['timestamp'],
            gift_id=data['giftId'],
            gift_type=data['giftType'],
            action=data['action'],
            price=data['price'],
            rnd=data['rnd'],
            coin_type=data['coin_type'],
            total_coin=data['total_coin'],
            tid=data['tid'],
        )


@dataclasses.dataclass
class GuardBuyMessage:
    """
    上舰消息
    """

    uid: int = 0
    """用户ID"""
    username: str = ''
    """用户名"""
    guard_level: int = 0
    """舰队等级，0非舰队，1总督，2提督，3舰长"""
    num: int = 0
    """数量"""
    price: int = 0
    """单价金瓜子数"""
    gift_id: int = 0
    """礼物ID"""
    gift_name: str = ''
    """礼物名"""
    start_time: int = 0
    """开始时间戳，和结束时间戳相同"""
    end_time: int = 0
    """结束时间戳，和开始时间戳相同"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            username=data['username'],
            guard_level=data['guard_level'],
            num=data['num'],
            price=data['price'],
            gift_id=data['gift_id'],
            gift_name=data['gift_name'],
            start_time=data['start_time'],
            end_time=data['end_time'],
        )


@dataclasses.dataclass
class SuperChatMessage:
    """
    醒目留言消息
    """

    price: int = 0
    """价格（人民币）"""
    message: str = ''
    """消息"""
    message_trans: str = ''
    """消息日文翻译（目前只出现在SUPER_CHAT_MESSAGE_JPN）"""
    start_time: int = 0
    """开始时间戳"""
    end_time: int = 0
    """结束时间戳"""
    time: int = 0
    """剩余时间（约等于 结束时间戳 - 开始时间戳）"""
    id: int = 0
    """醒目留言ID，删除时用"""
    gift_id: int = 0
    """礼物ID"""
    gift_name: str = ''
    """礼物名"""
    uid: int = 0
    """用户ID"""
    uname: str = ''
    """用户名"""
    face: str = ''
    """用户头像URL"""
    guard_level: int = 0
    """舰队等级，0非舰队，1总督，2提督，3舰长"""
    user_level: int = 0
    """用户等级"""
    background_bottom_color: str = ''
    """底部背景色，'#rrggbb'"""
    background_color: str = ''
    """背景色，'#rrggbb'"""
    background_icon: str = ''
    """背景图标"""
    background_image: str = ''
    """背景图URL"""
    background_price_color: str = ''
    """背景价格颜色，'#rrggbb'"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            price=data['price'],
            message=data['message'],
            message_trans=data['message_trans'],
            start_time=data['start_time'],
            end_time=data['end_time'],
            time=data['time'],
            id=data['id'],
            gift_id=data['gift']['gift_id'],
            gift_name=data['gift']['gift_name'],
            uid=data['uid'],
            uname=data['user_info']['uname'],
            face=data['user_info']['face'],
            guard_level=data['user_info']['guard_level'],
            user_level=data['user_info']['user_level'],
            background_bottom_color=data['background_bottom_color'],
            background_color=data['background_color'],
            background_icon=data['background_icon'],
            background_image=data['background_image'],
            background_price_color=data['background_price_color'],
        )


@dataclasses.dataclass
class SuperChatDeleteMessage:
    """
    删除醒目留言消息
    """

    ids: List[int] = dataclasses.field(default_factory=list)
    """醒目留言ID数组"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            ids=data['ids'],
        )


@dataclasses.dataclass
class InteractWordMessage:
    """
    交互事件
    """

    uid: int = 0
    """用户ID"""
    uname: str = ''
    """用户名"""
    msg_type: int = 0
    """消息类型"""
    timestamp: int = 0
    """时间戳"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            uname=data['uname'],
            msg_type=data['msg_type'],
            timestamp=data['timestamp'],
        )


@dataclasses.dataclass
class UserToastMessage:
    """
    用户大航海信息，用来代替 GuardBuyMessage
    """

    uid: int = 0
    """用户ID"""
    username: str = ''
    """用户名"""
    guard_level: int = 0
    """舰队等级，0非舰队，1总督，2提督，3舰长"""
    num: int = 0
    """数量"""
    price: int = 0
    """单价金瓜子数"""
    role_name: str = ''
    """舰长"""
    toast_msg: str = ''
    """`<%_username_%>` 自动续费了舰长"""
    start_time: int = 0
    """开始时间戳，和结束时间戳相同"""
    end_time: int = 0
    """结束时间戳，和开始时间戳相同"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            username=data['username'],
            guard_level=data['guard_level'],
            num=data['num'],
            price=data['price'],
            role_name=data['role_name'],
            toast_msg=data['toast_msg'],
            start_time=data['start_time'],
            end_time=data['end_time'],
        )


@dataclasses.dataclass
class UserVirtualMvp:
    """
    守护圣法师事件
    """

    action: str = ''
    """文本「解锁」"""
    goods_icon: str = ''
    """礼物图标URL"""
    goods_name: str = ''
    """「守护圣法师x7天」"""
    goods_num: int = 0
    """数量"""
    goods_price: int = 0
    """价格，要除以 1000"""
    success_toast: str = ''
    """「解锁成功，已为您穿戴守护圣法师装扮」"""
    timestamp: int = 0
    """时间戳"""
    uid: int = 0
    """用户ID"""
    uname: str = ''
    """用户名"""
    user_guard_level: int = 0
    """舰队等级，0非舰队，1总督，2提督，3舰长"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            uname=data['uname'],
            user_guard_level=data['user_guard_level'],
            goods_num=data['goods_num'],
            goods_price=data['goods_price'],
            success_toast=data['success_toast'],
            action=data['action'],
            goods_icon=data['goods_icon'],
            goods_name=data['goods_name'],
            timestamp=data['timestamp'],
        )

@dataclasses.dataclass
class WarningMessage:
    """
    直播间被警告
    """

    roomid: int = 0
    """直播间id"""
    msg: str = ''
    """警告原因，ie. 图片内容不适宜，请立即调整"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            roomid=data['roomid'],
            msg=data['msg'],
        )

@dataclasses.dataclass
class CutOffMessage:
    """
    直播间被切断
    """

    roomid: int = 0
    """直播间id"""
    msg: str = ''
    """切断原因，ie. 违反直播言论规范，请立即调整"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            roomid=data['roomid'],
            msg=data['msg'],
        )

@dataclasses.dataclass
class RoomSilentOnMessage:
    """
    开启等级禁言
    """

    type: str = ''
    """禁言类型，level 用户等级禁言，medal 粉丝勋章等级禁言，member 全员"""
    level: int = 0
    """禁言等级"""
    second: int = 0
    """时间戳"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            type=data['type'],
            level=data['level'],
            second=data['second'],
        )


@dataclasses.dataclass
class RoomSilentOffMessage:
    """
    关闭等级禁言
    """
    type: str = ''
    """禁言类型，关闭禁言时返回 ''"""
    level: int = 0
    """禁言等级，关闭禁言时返回 0"""
    second: int = 0
    """时间戳"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            type=data['type'],
            level=data['level'],
            second=data['second'],
        )


@dataclasses.dataclass
class RoomAdminsMessage:
    """
    房管列表变更
    """
    uids: str = ''
    """当前房管UIDs"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uids=data['uids'],
        )


@dataclasses.dataclass
class RoomAdminEntranceMessage:
    """
    任命房管
    """

    uid: int = 0
    """用户ID"""
    dmscore: int = 0
    """45，作用未知"""
    level: int = 0
    """1"""
    msg: str = ''
    """统提示：你已被主播设为房管"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            dmscore=data['dmscore'],
            level=data['level'],
            msg=data['msg'],
        )


@dataclasses.dataclass
class RoomAdminRevokeMessage:
    """
    撤销房管
    """

    uid: int = 0
    """被撤销房管的uid"""
    msg: str = ''
    """文本，撤销房管"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            msg=data['msg'],
        )


@dataclasses.dataclass
class RoomBlockMessage:
    """
    指定观众禁言
    """

    uid: int = 0
    """被拉黑用户的uid"""
    uname: str = ''
    """被拉黑用户的用户名"""
    operator: int = 0
    """拉黑操作者，1为房管，2为主播，其它未知"""
    dmscore: int = 0
    """30，作用未知"""

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            uid=data['uid'],
            uname=data['uname'],
            operator=data['operator'],
            dmscore=data['dmscore'],
        )
